var __defProp = Object.defineProperty;
var __defProps = Object.defineProperties;
var __getOwnPropDescs = Object.getOwnPropertyDescriptors;
var __getOwnPropSymbols = Object.getOwnPropertySymbols;
var __hasOwnProp = Object.prototype.hasOwnProperty;
var __propIsEnum = Object.prototype.propertyIsEnumerable;
var __defNormalProp = (obj, key, value) => key in obj ? __defProp(obj, key, { enumerable: true, configurable: true, writable: true, value }) : obj[key] = value;
var __spreadValues = (a, b) => {
  for (var prop in b || (b = {}))
    if (__hasOwnProp.call(b, prop))
      __defNormalProp(a, prop, b[prop]);
  if (__getOwnPropSymbols)
    for (var prop of __getOwnPropSymbols(b)) {
      if (__propIsEnum.call(b, prop))
        __defNormalProp(a, prop, b[prop]);
    }
  return a;
};
var __spreadProps = (a, b) => __defProps(a, __getOwnPropDescs(b));

// src/transformer/transformer.ts
import traverse3 from "@babel/traverse";

// src/transformer/impl/state.ts
function createState() {
  const state = new Map();
  return {
    has(key) {
      return state.has(key);
    },
    get(key) {
      return state.get(key);
    },
    getOrSet(key, or) {
      const v = state.get(key);
      if (v !== void 0)
        return v;
      state.set(key, or);
      return or;
    },
    set(key, value) {
      state.set(key, value);
      return value;
    },
    delete(key) {
      state.delete(key);
    }
  };
}

// src/transformer/transformer.ts
import { parse as parse2 } from "@babel/parser";

// src/transformer/traverse.ts
import {
  isIdentifier,
  isImportDefaultSpecifier,
  isImportNamespaceSpecifier,
  isMemberExpression,
  isStringLiteral
} from "@babel/types";
import traverse from "@babel/traverse";
var nodeLoc = (node) => node.loc ? `line ${node.loc.start.line}, column ${node.loc.start.column}` : "unknown";
function findProgramPath(ast) {
  let path;
  traverse(ast, {
    Program(p) {
      path = p;
      p.stop();
    }
  });
  return path;
}
function isLocalDefined(scope, name) {
  return scope.getBinding(name);
}
function createImportedMacroContainer(macros, container = []) {
  return {
    size: () => container.length,
    collectFromAST,
    collectFromNodePath,
    getCalledMacro,
    testContainsMacros
  };
  function collectFromAST(ast) {
    const handler = (path) => collectFromNodePath(path);
    traverse(ast, {
      ImportDeclaration(path) {
        handler(path);
      }
    });
    findProgramPath(ast).scope.crawl();
  }
  function collectFromNodePath(path) {
    const moduleName = path.node.source.value;
    if (!(moduleName in macros))
      return;
    path.node.specifiers.forEach((s) => {
      if (isImportDefaultSpecifier(s) || isImportNamespaceSpecifier(s)) {
        container.push({
          importedAsNamespace: true,
          moduleName: path.node.source.value,
          local: s.local.name,
          macros: macros[moduleName]
        });
      } else {
        const exportName = isIdentifier(s.imported) ? s.imported.name : s.imported.value;
        container.push({
          importedAsNamespace: false,
          moduleName: path.node.source.value,
          local: s.local.name,
          macro: macros[moduleName].find((m) => m.name === exportName)
        });
      }
    });
    path.remove();
  }
  function getCalledMacro(callee) {
    const node = callee.node;
    if (isMemberExpression(node)) {
      const ns = node.object;
      if (isIdentifier(ns)) {
        if (isLocalDefined(callee.scope, ns.name))
          return;
        const maybeMacro = container.find((m) => m.importedAsNamespace && m.local === ns.name);
        if (!maybeMacro)
          return;
        const method = node.property;
        if (isIdentifier(method)) {
          return maybeMacro.macros.find((m) => m.name === method.name) || method.name;
        } else if (isStringLiteral(method)) {
          return maybeMacro.macros.find((m) => m.name === method.value) || method.value;
        }
      }
    } else if (isIdentifier(node)) {
      if (isLocalDefined(callee.scope, node.name))
        return;
      const maybeMacro = container.find((m) => !m.importedAsNamespace && m.local === node.name);
      if (!maybeMacro)
        return;
      return maybeMacro.macro || node.name;
    }
  }
  function testContainsMacros(paths) {
    return paths.map((path) => {
      if (path.isCallExpression() && getCalledMacro(path.get("callee")))
        return true;
      let has = false;
      path.traverse({
        CallExpression(p) {
          if (getCalledMacro(p.get("callee"))) {
            has = true;
            p.stop();
          }
        }
      });
      return has;
    });
  }
}

// src/transformer/transformer.ts
import generate2 from "@babel/generator";

// src/transformer/impl/babel.ts
import traverse2 from "@babel/traverse";
import * as types from "@babel/types";
import template from "@babel/template";
import { parse, parseExpression } from "@babel/parser";
import generate from "@babel/generator";
var BABEL = Object.freeze({
  template,
  traverse: traverse2,
  types,
  parse,
  parseExpression,
  generate
});

// src/transformer/impl/import.ts
import {
  isIdentifier as isIdentifier2,
  isImportDefaultSpecifier as isImportDefaultSpecifier2,
  isImportNamespaceSpecifier as isImportNamespaceSpecifier2,
  isImportSpecifier
} from "@babel/types";
var renderImportStmt = (imp) => "defaultName" in imp ? `import ${imp.defaultName} from '${imp.moduleName}'` : "localName" in imp ? `import { ${imp.exportName} as ${imp.localName} } from '${imp.moduleName}'` : "exportName" in imp ? `import { ${imp.exportName} } from '${imp.moduleName}'` : "namespaceName" in imp ? `import * as ${imp.namespaceName} from '${imp.moduleName}'` : `import '${imp.moduleName}'`;
function matchImportStmt(imp, node, loose) {
  if (node.source.value !== imp.moduleName)
    return false;
  if ("defaultName" in imp) {
    for (const s of node.specifiers) {
      if (isImportDefaultSpecifier2(s) && s.local.name === imp.defaultName)
        return true;
    }
  } else if ("localName" in imp) {
    for (const s of node.specifiers) {
      if (isImportSpecifier(s) && s.local.name === imp.localName && isIdentifier2(s.imported) && s.imported.name === imp.exportName)
        return true;
    }
  } else if ("exportName" in imp) {
    for (const s of node.specifiers) {
      if (isImportSpecifier(s) && isIdentifier2(s.imported) && s.imported.name === imp.exportName && (loose || s.local.name === imp.exportName))
        return true;
    }
  } else if ("namespaceName" in imp) {
    for (const s of node.specifiers) {
      if (isImportNamespaceSpecifier2(s) && s.local.name === imp.namespaceName)
        return true;
    }
  } else {
    if (loose || node.specifiers.length === 0)
      return true;
  }
  return false;
}

// src/transformer/impl/macroHelper.ts
import template2 from "@babel/template";
import { isArray } from "@typed-macro/shared";
function createMacroHelper(thisProgram, importedMacros) {
  return {
    findImported: (imp, loose = false, program = thisProgram) => {
      if (isArray(imp))
        return imp.map((i) => findImported(program, i, loose));
      return findImported(program, imp, loose);
    },
    prependImports: (imports, program = thisProgram) => prependImports(program, imports),
    appendImports: (imports, program = thisProgram) => appendImports(program, imports),
    prependToBody: (nodes, program = thisProgram) => {
      return prependToBody(program, nodes);
    },
    appendToBody: (nodes, program = thisProgram) => {
      return appendToBody(program, nodes);
    },
    containsMacros: (paths) => {
      if (!isArray(paths))
        return importedMacros.testContainsMacros([paths])[0];
      return importedMacros.testContainsMacros(paths);
    }
  };
}
function findImported(program, imp, loose) {
  for (const path of program.get("body")) {
    if (path.isImportDeclaration() && matchImportStmt(imp, path.node, loose))
      return path;
  }
}
function prependImports(program, imports) {
  let lastInserted;
  const inserter = (stmt) => lastInserted = lastInserted ? lastInserted.insertAfter(stmt)[0] : program.unshiftContainer("body", stmt)[0];
  return insertImports(program, imports, inserter);
}
function appendImports(program, imports) {
  let lastInserted = program.get("body").filter((p) => p.isImportDeclaration()).pop();
  const inserter = (stmt) => lastInserted = lastInserted ? lastInserted.insertAfter(stmt)[0] : program.unshiftContainer("body", stmt)[0];
  return insertImports(program, imports, inserter);
}
function prependToBody(program, nodes) {
  if (isArray(nodes))
    return program.unshiftContainer("body", nodes);
  return program.unshiftContainer("body", nodes)[0];
}
function appendToBody(program, nodes) {
  if (isArray(nodes))
    return program.pushContainer("body", nodes);
  return program.pushContainer("body", nodes)[0];
}
function insertImports(program, imports, inserter) {
  if (isArray(imports)) {
    if (!imports.length)
      return [];
    return imports.map((imp) => {
      const existed2 = findImported(program, imp, false);
      if (existed2)
        return existed2;
      return inserter(template2.statement.ast(renderImportStmt(imp)));
    });
  }
  const existed = findImported(program, imports, false);
  if (existed)
    return existed;
  return inserter(template2.statement.ast(renderImportStmt(imports)));
}

// src/transformer/transformer.ts
import { isArray as isArray2 } from "@typed-macro/shared";
function createTransformer(options) {
  const macros = Object.create(null);
  const appendMacros = (moduleName, ms) => {
    const m = macros[moduleName] || (macros[moduleName] = []);
    m.push(...ms);
    checkMacrosDuplicated(moduleName, m);
  };
  let parserPlugins = ["typescript", "jsx"];
  const appendParserPlugins = (plugins) => {
    parserPlugins = Array.from(new Set([...parserPlugins, ...plugins]));
  };
  const maxTraversals = options.maxTraversals && options.maxTraversals > 1 ? options.maxTraversals : 5;
  if (options.parserPlugins)
    appendParserPlugins(options.parserPlugins);
  return {
    appendMacros,
    appendParserPlugins,
    transform: (code, filepath, env) => {
      const ast = parse2(code, {
        sourceType: "module",
        plugins: parserPlugins
      });
      const importedMacros = createImportedMacroContainer(macros);
      importedMacros.collectFromAST(ast);
      if (!importedMacros.size())
        return;
      const transformState = createState();
      let traversalTimes = 0;
      while (applyMacros()) {
        importedMacros.collectFromAST(ast);
        if (traversalTimes++ >= maxTraversals)
          throw new Error(`Reached the maximum traversal times (${maxTraversals}) when apply macros on '${filepath}'`);
      }
      return generate2(ast, {
        retainLines: true
      }).code;
      function applyMacros() {
        let applied = false;
        const program = findProgramPath(ast);
        const traversalState = createState();
        traverse3(ast, {
          CallExpression(path) {
            if (process(path))
              applied = true;
          }
        });
        return applied;
        function handleYield(current, ctx) {
          if (!ctx)
            return;
          if (!isArray2(ctx))
            ctx = [ctx];
          ctx.forEach((path) => {
            if (!path)
              return;
            if (path.isImportDeclaration()) {
              importedMacros.collectFromNodePath(path);
            } else {
              if (current.findParent((parent) => parent.node === path.node))
                throw new Error("can not yield a parent node");
              if (path.isCallExpression() && process(path))
                return;
              path.traverse({
                CallExpression(p) {
                  process(p);
                }
              });
            }
          });
        }
        function process(path) {
          const macro = importedMacros.getCalledMacro(path.get("callee"));
          if (!macro)
            return false;
          if (typeof macro === "string")
            throw new Error(`Macro '${macro}' is not existed but called in '${filepath}' near ${nodeLoc(path.node)}`);
          try {
            const transform = macro.handler({
              ctx: __spreadProps(__spreadValues({}, env), {
                filepath,
                code,
                ast,
                path,
                args: path.get("arguments"),
                state: {
                  transform: transformState,
                  traversal: traversalState
                }
              }),
              babel: BABEL,
              helper: createMacroHelper(program, importedMacros)
            });
            for (const ctx of transform)
              handleYield(path, ctx);
          } catch (e) {
            throw new Error(`Error when apply macro '${macro.name}' in '${filepath}' near ${nodeLoc(path.node)}:
 ${e}`);
          }
          return true;
        }
      }
    }
  };
}
function checkMacrosDuplicated(moduleName, macros) {
  const mem = Object.create(null);
  macros.forEach((m) => {
    if (mem[m.name])
      throw new Error(`macro '${m.name}' in '${moduleName}' is duplicated`);
    mem[m.name] = 1;
  });
}

// src/normalizer/normalizer.ts
import {
  isMacro,
  isMacroProvider
} from "@typed-macro/core";
import { isFunction, VERSION } from "@typed-macro/shared";
function normalizeExports(exports) {
  const macros = Object.create(null);
  const codes = Object.create(null);
  const types2 = Object.create(null);
  Object.keys(exports).forEach((moduleName) => {
    const item = exports[moduleName];
    const type = types2[moduleName] || (types2[moduleName] = []);
    if (item.types)
      type.push(item.types);
    if ("code" in item) {
      codes[moduleName] = (codes[moduleName] || "") + (item.code || "") + "\n";
    } else {
      item.macros.forEach((m) => validateMacro(moduleName, m));
      const ms = macros[moduleName] || (macros[moduleName] = []);
      ms.push(...item.macros);
      type.push(...item.macros.map((m) => m.types));
    }
  });
  return {
    macros,
    codes,
    types: types2
  };
}
function normalizeMacroProvider(provider, env) {
  validateMacroProvider(provider);
  const instantiated = isFunction(provider) ? provider(env) : provider;
  instantiated.normalizedExports = normalizeExports(instantiated.exports);
  return instantiated;
}
function validateMacro(moduleName, m) {
  if (!isMacro(m))
    throw new Error(`'${m}' in '${moduleName}' is not a macro`);
  if (m.version !== VERSION)
    throw new Error(`macro '${m.name}' in '${moduleName}' is incompatible, expect '${VERSION}', actually '${m.version}'`);
}
function validateMacroProvider(provider) {
  if (!isMacroProvider(provider))
    throw new Error(`not a macro provider`);
  if (provider.version !== VERSION)
    throw new Error(`incompatible version, expect '${VERSION}', actually '${provider.version}'`);
}

// src/renderer/renderer.ts
import { mkdir, writeFile } from "fs/promises";
import { dirname } from "path";
function createTypeRenderer() {
  const types2 = Object.create(null);
  return {
    renderToString: () => renderTypes(types2),
    renderToFile: async (path) => {
      try {
        await mkdir(dirname(path), { recursive: true });
        await writeFile(path, renderTypes(types2));
      } catch (e) {
        throw new Error(`Error when write type declaration file: ${e}`);
      }
    },
    append: (moduleName, ts) => {
      const t = types2[moduleName] || (types2[moduleName] = []);
      t.push(...ts);
    }
  };
}
function renderTypes(types2) {
  const moduleNames = Object.keys(types2);
  return moduleNames.map((moduleName) => {
    const item = types2[moduleName];
    return `declare module '${moduleName}' {
${item.join("\n")}
}`;
  }).join("\n");
}

// src/filter/filter.ts
import pm from "picomatch";
import { isArray as isArray3, normalizePath } from "@typed-macro/shared";
function createFilter({
  include = /\.[jt]sx?$/,
  exclude = /node_modules/
}) {
  const getMatcher = (pattern) => {
    if (pattern instanceof RegExp)
      return pattern;
    const fn = pm(pattern, { dot: true });
    return { test: (id) => fn(id) };
  };
  const ensureArray = (v) => v ? isArray3(v) ? v : [v] : [];
  const includeMatchers = ensureArray(include).map(getMatcher);
  const excludeMatchers = ensureArray(exclude).map(getMatcher);
  return {
    isExcluded(id) {
      const path = normalizePath(id);
      return !!excludeMatchers.find((pattern) => pattern.test(path));
    },
    isIncluded(id) {
      const path = normalizePath(id);
      return !!includeMatchers.find((pattern) => pattern.test(path));
    }
  };
}

// src/loader/loader.ts
function createLoader() {
  const modules = Object.create(null);
  return {
    append: (moduleName, code) => {
      if (modules[moduleName]) {
        throw new Error(`duplicated virtual module '${moduleName}'`);
      }
      if (code === void 0)
        modules[moduleName] = "export {}";
      else
        modules[moduleName] = code;
    },
    load: (moduleName) => modules[moduleName]
  };
}

// src/hook/hook.ts
import { isString } from "@typed-macro/shared";
function createHookContainer() {
  const container = {
    onStart: [],
    onFilter: [],
    beforeTransform: [],
    afterTransform: [],
    onStop: []
  };
  return {
    append: (hooks) => {
      const keys = Object.keys(hooks);
      keys.forEach((key) => {
        if (!(key in container))
          return;
        container[key].push(hooks[key]);
      });
    },
    onStart: async () => {
      await Promise.all(container.onStart.map((hook) => hook()));
    },
    onStop: async () => {
      await Promise.all(container.onStop.map((hook) => hook()));
    },
    onFilter: async (id) => {
      for (const hook of container.onFilter) {
        if (await hook(id))
          return true;
      }
      return false;
    },
    beforeTransform: createTransformHook(container.beforeTransform),
    afterTransform: createTransformHook(container.afterTransform)
  };
}
function createTransformHook(hooks) {
  return async (code, id) => {
    for (const hook of hooks) {
      const result = await hook(code, id);
      if (result != null) {
        if (isString(result))
          code = result;
        else
          code = result.code || code;
      }
    }
    return code;
  };
}

// src/runtime/runtime.ts
function createRuntime(env, options) {
  const filter = createFilter(options.filter || {});
  const loader = createLoader();
  const renderer = createTypeRenderer();
  const transformer = createTransformer(options.transformer || {});
  const hookContainer = createHookContainer();
  let providerCount = 0;
  return {
    appendProvider(provider) {
      try {
        const normalized = normalizeMacroProvider(provider, env);
        appendExports(normalized.normalizedExports);
        normalized.hooks && appendHooks(normalized.hooks);
        normalized.options && appendOptions(normalized.options);
        providerCount++;
      } catch (e) {
        throw new Error(`Error when add the macro provider at index ${providerCount}: ${e}`);
      }
    },
    async renderTypes(filepath) {
      await renderer.renderToFile(filepath);
    },
    async start() {
      await hookContainer.onStart();
    },
    async stop() {
      await hookContainer.onStop();
    },
    async transform(code, id) {
      const preTransformed = await hookContainer.beforeTransform(code, id);
      const transformed = transformer.transform(preTransformed, id, env);
      return await hookContainer.afterTransform(transformed || preTransformed, id);
    },
    async filter(id) {
      if (filter.isExcluded(id))
        return false;
      return filter.isIncluded(id) || await hookContainer.onFilter(id);
    },
    load(id) {
      return loader.load(id);
    },
    resolveId(id) {
      return loader.load(id) ? id : void 0;
    },
    internal: {
      env,
      filter,
      loader,
      renderer,
      transformer,
      hookContainer
    }
  };
  function appendExports({ types: types2, codes, macros }) {
    Object.keys(types2).forEach((moduleName) => {
      renderer.append(moduleName, types2[moduleName]);
    });
    Object.keys(codes).forEach((moduleName) => {
      loader.append(moduleName, codes[moduleName]);
    });
    Object.keys(macros).forEach((moduleName) => {
      loader.append(moduleName);
      transformer.appendMacros(moduleName, macros[moduleName]);
    });
  }
  function appendHooks(hooks) {
    hookContainer.append(hooks);
  }
  function appendOptions(options2) {
    options2.parserPlugins && transformer.appendParserPlugins(options2.parserPlugins);
  }
}
export {
  createFilter,
  createHookContainer,
  createLoader,
  createRuntime,
  createTransformer,
  createTypeRenderer,
  normalizeExports,
  normalizeMacroProvider
};
